@@ -3,40 +3,45 @@ import hashlib
 import logging
 import os
 import asyncio
+import random
 import socket
 import struct
+import time
 import typing
 
 import requests
 
 from .sync import persistent_state_ttl, choose_key_block, sync
 from .utils import init_mainnet_block, init_testnet_block
-from ..boc import Slice, Cell, Builder
-from ..proof.check_proof import check_block_header_proof, check_shard_proof, check_account_proof, check_proof, \
+from pytoniq_core.boc import Slice, Cell, Builder
+from pytoniq_core.proof.check_proof import check_block_header_proof, check_shard_proof, check_account_proof, check_proof, \
     check_block_signatures, compute_validator_set
-from ..boc.address import Address
+from pytoniq_core.boc.address import Address
 
-# from .crypto import ed25519Public, ed25519Private, x25519Public, x25519Private
-from ..crypto.ciphers import Server, Client, get_random, create_aes_ctr_cipher, aes_ctr_encrypt, aes_ctr_decrypt, get_shared_key
-from ..crypto.crc import crc16
+from pytoniq_core.crypto.ciphers import Server, Client, get_random, create_aes_ctr_cipher, aes_ctr_encrypt, aes_ctr_decrypt, get_shared_key
+from pytoniq_core.crypto.crc import crc16
 
-from ..tl.generator import TlGenerator, TlSchema
-from ..tl.block import BlockId, BlockIdExt
-from ..tlb.config import ConfigParam34, ConfigParam28, ConfigParam
-from ..tlb.transaction import Transaction
-from ..tlb.utils import deserialize_shard_hashes
+from pytoniq_core.tl.generator import TlGenerator, TlSchema
+from pytoniq_core.tl.block import BlockId, BlockIdExt
+from pytoniq_core.tlb.config import ConfigParam34, ConfigParam28, ConfigParam
+from pytoniq_core.tlb.transaction import Transaction
+from pytoniq_core.tlb.utils import deserialize_shard_hashes
 
-from ..tlb.vm_stack import VmStack
-from ..tlb.block import Block, ShardDescr, BinTree, ShardStateUnsplit, KeyExtBlkRef
-from ..tlb.account import Account, SimpleAccount, ShardAccount, AccountBlock
+from pytoniq_core.tlb.vm_stack import VmStack
+from pytoniq_core.tlb.block import Block, ShardDescr, BinTree, ShardStateUnsplit, KeyExtBlkRef
+from pytoniq_core.tlb.account import Account, SimpleAccount, ShardAccount, AccountBlock
 
 
-class LiteClientError(BaseException):
+class LiteClientError(Exception):
     pass
 
 
 class RunGetMethodError(LiteClientError):
-    pass
+    def __init__(self, address: typing.Any, method: typing.Any, exit_code: int):
+        self.address = address
+        self.method = method
+        self.exit_code = exit_code
+        super().__init__(f'get method "{method}" for account {address} returned exit code {exit_code}')
 
 
 class LiteClient:
@@ -45,15 +50,20 @@ class LiteClient:
                  host: str,  # ipv4 host
                  port: int,
                  server_pub_key: str,  # server ed25519 public key in base64,
+                 timeout: int = 10,
                  tl_schemas_path: typing.Optional[str] = None,
                  trust_level: int = 1,
                  init_key_block: BlockIdExt = None,
                  ) -> None:
+        """
+        ADNL over TCP client for `liteservers` usage
+        """
 
         """########### init ###########"""
         self.tasks = {}
         self.inited = False
         self.logger = logging.getLogger(self.__class__.__name__)
+        self.timeout = timeout
 
         """########### sync ###########"""
         self.last_mc_block: BlockIdExt = None
@@ -73,16 +83,22 @@ class LiteClient:
         """########### connection ###########"""
         self.reader: asyncio.StreamReader = None
         self.writer: asyncio.StreamWriter = None
+        self.loop: asyncio.AbstractEventLoop = None
+        self.delta = 0.02  # listen delay
+
         self.listener: asyncio.Task = None
+        rand_int = str(random.randint(10**9, 3 * 10**9))
+        self.listener_name: str = 'listener' + rand_int
         self.pinger: asyncio.Task = None
+        self.pinger_name: str = 'pinger' + rand_int
         self.updater: asyncio.Task = None
-        self.loop = asyncio.get_event_loop()
-        self.delta = 0.1  # listen delay
+        self.updater_name: str = 'updater' + rand_int
 
         """########### TL ###########"""
         if tl_schemas_path is None:
-            tl_schemas_path = os.path.join(os.path.dirname(__file__), os.pardir, 'tl/schemas')
-        self.schemas = TlGenerator(tl_schemas_path).generate()
+            self.schemas = TlGenerator.with_default_schemas().generate()
+        else:
+            self.schemas = TlGenerator(tl_schemas_path).generate()
         # for better performance:
         self.ping_sch = self.schemas.get_by_name('tcp.ping')
         self.pong_sch = self.schemas.get_by_name('tcp.pong')
@@ -136,28 +152,33 @@ class LiteClient:
             if not result:
                 # for handshake
                 result = {}
-            if 'code' in result and 'message' in result:
-                raise LiteClientError(f'LiteClient crashed with {result["code"]} code. Message: {result["message"]}')
 
             qid = result.get('query_id', result.get('random_id'))  # return query_id for ordinary requests, random_id for ping-pong requests, None for handshake
 
             request = self.tasks.pop(qid)
-            request.set_result(result.get('answer', {}))
+
+            result = result.get('answer', {})
+            request.set_result(result)
 
     async def connect(self) -> None:
+        self.loop = asyncio.get_running_loop()
         handshake = self.handshake()
-        self.reader, self.writer = await asyncio.open_connection(self.server.host, self.server.port)
-        future = await self.send(handshake, None)
-        self.listener = asyncio.create_task(self.listen(), name='listener')
+        self.reader, self.writer = await asyncio.wait_for(asyncio.open_connection(self.server.host, self.server.port), self.timeout)
+        future = await asyncio.wait_for(self.send(handshake, None), self.timeout)
+        self.listener = asyncio.create_task(self.listen(), name=self.listener_name)
         await self.update_last_blocks()
-        self.pinger = asyncio.create_task(self.ping(), name='pinger')
-        self.updater = asyncio.create_task(self.block_updater(), name='updater')
+        self.pinger = asyncio.create_task(self.ping(), name=self.pinger_name)
+        self.updater = asyncio.create_task(self.block_updater(), name=self.updater_name)
         await future
         self.inited = True
 
+    async def reconnect(self) -> None:
+        await self.close()
+        await self.connect()
+
     async def close(self) -> None:
         for i in asyncio.all_tasks(self.loop):
-            if i.get_name() in ('pinger', 'listener', 'updater'):
+            if i.get_name() in {self.listener_name, self.pinger_name, self.updater_name}:
                 i.cancel()
         self.logger.info(msg='client has been closed')
 
@@ -215,14 +236,22 @@ class LiteClient:
             await pong
             self.logger.debug(msg=f'ping - pong')
 
+    async def liteserver_query(self, query: bytes, qid: str) -> dict:
+        data = self.serialize_packet(query)
+        resp = await self.send_and_encrypt(data, qid)
+        await asyncio.wait_for(resp, self.timeout)
+        result = resp.result()
+
+        if 'code' in result and 'message' in result:
+            raise LiteClientError(f'LiteClient crashed with {result["code"]} code. Message: {result["message"]}')
+
+        return resp.result()
+
     async def liteserver_request(self, tl_schema_name: str, data: dict) -> dict:
         schema = self.schemas.get_by_name('liteServer.' + tl_schema_name)
         self.logger.info(msg=f'requesting {tl_schema_name} with provided data {data}')
         data, qid = self.serialize_adnl_ls_query(schema, data)
-        data = self.serialize_packet(data)
-        resp = await self.send_and_encrypt(data, qid)
-        await resp
-        return resp.result()
+        return await self.liteserver_query(data, qid)
 
     @staticmethod
     def pack_block_id_ext(**kwargs):
@@ -258,16 +287,33 @@ class LiteClient:
         if self.last_mc_block is None:
             self.last_mc_block = await self.get_trusted_last_mc_block()
         while True:
-            result = await self.wait_masterchain_seqno(self.last_mc_block.seqno + 1, timeout_ms=10000)
-            if result['code'] != 0:
-                logging.getLogger().warning(f'error response from liteserver in block updater: {result}')
+            await self.wait_masterchain_seqno(self.last_mc_block.seqno + 1, timeout_ms=10000, schema_name='getMasterchainInfo', data={})
             await self.update_last_blocks()
 
     async def get_masterchain_info(self):
         return await self.liteserver_request('getMasterchainInfo', {})
 
-    async def wait_masterchain_seqno(self, seqno: int, timeout_ms: int):
-        return await self.liteserver_request('waitMasterchainSeqno', {'seqno': seqno, 'timeout_ms': timeout_ms})
+    async def raw_wait_masterchain_seqno(self, seqno: int, timeout_ms: int, suffix: bytes = b''):
+        prefix = self.schemas.serialize(schema=self.schemas.get_by_name('liteServer.waitMasterchainSeqno'), data={'seqno': seqno, 'timeout_ms': timeout_ms})
+
+        qid = get_random(32)
+        data = self.schemas.serialize(
+            self.adnl_query_sch,
+            {'query_id': qid,
+             'query': self.schemas.serialize(self.ls_query_sch,
+                                             {'data': prefix + suffix}
+                                             )
+             }
+        )
+        return await self.liteserver_query(data, qid[::-1].hex())
+
+    async def wait_masterchain_seqno(self, seqno: int, timeout_ms: int, schema_name: str, data: dict = None):
+        if data is None:
+            data = {}
+
+        suffix = self.schemas.serialize(self.schemas.get_by_name('liteServer.' + schema_name), data)
+
+        return await self.raw_wait_masterchain_seqno(seqno, timeout_ms, suffix)
 
     async def get_masterchain_info_ext(self):
         return await self.liteserver_request('getMasterchainInfoExt', {'mode': 0})
@@ -359,7 +405,6 @@ class LiteClient:
 
         data = {'id': block.to_dict(), 'account': account}
         result = await self.liteserver_request('getAccountState', data)
-
         shrd_blk = BlockIdExt.from_dict(result['shardblk'])
         if not result['state']:
             return None, None  # account_none$0 = Account;
@@ -407,7 +452,7 @@ class LiteClient:
         result = await self.liteserver_request('runSmcMethod', data)
 
         if result['exit_code'] != 0:
-            raise RunGetMethodError(f'get method "{method}" for account {address} returned exit code {result["exit_code"]}')
+            raise RunGetMethodError(address=address, method=method, exit_code=result['exit_code'])
         if self.trust_level <= 1:
             shrd_blk = BlockIdExt.from_dict(result['shardblk'])
             check_shard_proof(shard_proof=result['shard_proof'], blk=block, shrd_blk=shrd_blk)
@@ -471,6 +516,28 @@ class LiteClient:
 
         return deserialize_shard_hashes(shard_hashes_cell.begin_parse())
 
+    async def get_all_shards_info(self, block: typing.Optional[BlockIdExt] = None) -> typing.List[BlockIdExt]:
+        """
+        High-level function to get shards as `BlockIdExt` list instead of BinTrees
+        :param block: block to get shards of
+        :return: list of BlockIdExt of all block shards
+        """
+        shards = await self.raw_get_all_shards_info(block)
+        result = []
+        for k, v in shards.items():
+            for sh in v.list:
+                sh: ShardDescr
+                result.append(
+                    BlockIdExt(
+                        workchain=k,
+                        shard=sh.next_validator_shard_signed,
+                        seqno=sh.seq_no,
+                        root_hash=sh.root_hash,
+                        file_hash=sh.file_hash
+                    )
+                )
+        return result
+
     async def get_one_transaction(self, address: typing.Union[Address, str], lt: int, block: BlockIdExt) -> typing.Optional[Transaction]:
         if isinstance(address, str):
             address = Address(address)
@@ -578,9 +645,11 @@ class LiteClient:
         data = {'id': block.to_dict(), 'mode': mode, 'count': count, 'want_proof': b''}
         result = await self.liteserver_request('listBlockTransactionsExt', data)
 
-        transactions_cells = Cell.from_boc(result['transactions'])
         if not result['transactions']:
             return []
+
+        transactions_cells = Cell.from_boc(result['transactions'])
+
         if self.trust_level <= 1:
             proof = Cell.one_from_boc(result['proof'])
             check_block_header_proof(proof[0], block.root_hash)
@@ -592,7 +661,7 @@ class LiteClient:
         for tr_root in transactions_cells:
             transaction = Transaction.deserialize(tr_root.begin_parse())
             if self.trust_level <= 1:
-                prunned_tr_cell = acc_block.get(int(transaction.account_addr, 16)).transactions[0].get(transaction.lt)
+                prunned_tr_cell = acc_block.get(int(transaction.account_addr_hex, 16)).transactions[0].get(transaction.lt)
                 assert prunned_tr_cell.get_hash(0) == tr_root.get_hash(0)
             # transaction.account = Address((block.workchain, bytes.fromhex(transaction.account_addr)))
             tr_result.append(transaction)
@@ -767,17 +836,40 @@ class LiteClient:
         if prove_mc:
             await self.get_mc_block_proof(known_block=self.last_key_block, target_block=mc_block)
 
+        def check_shard_in_master(proof: Cell, blk: BlockIdExt):
+            check_block_header_proof(proof[0], mc_block.root_hash)
+
+            shards = Block.deserialize(proof[0].begin_parse()).extra.custom.shard_hashes[blk.workchain].list
+            shard = None
+            for sh in shards:
+                sh: ShardDescr
+                if sh.seq_no == blk.seqno and sh.next_validator_shard_signed == blk.shard:
+                    shard = sh.__dict__
+
+            shardblk = BlockIdExt.from_dict(shard)
+            shardblk.seqno = shard['seq_no']
+            shardblk.workchain = blk.workchain
+            return shardblk
+
+        if len(result['links']) == 1:
+            assert check_shard_in_master(Cell.one_from_boc(result['links'][0]['proof']), blk) == blk
+            return
+
+        last_shard_blk = None
+
         for link in result['links']:
-            if BlockIdExt.from_dict(link['id']) == blk:
-                proof = Cell.one_from_boc(link['proof'])
-                check_block_header_proof(proof[0], mc_block.root_hash)
-                shard = Block.deserialize(proof[0].begin_parse()).extra.custom.shard_hashes[blk.workchain].list[0].__dict__
-                shardblk = BlockIdExt.from_dict(shard)
-                shardblk.seqno = shard['seq_no']
-                shardblk.workchain = blk.workchain
-                assert shardblk == blk
-            else:
-                raise NotImplementedError()
+            proof = Cell.one_from_boc(link['proof'])
+            if proof[0].get_hash(0) == mc_block.root_hash:
+                last_shard_blk = check_shard_in_master(proof, BlockIdExt.from_dict(link['id']))
+                continue
+            check_block_header_proof(proof[0], last_shard_blk.root_hash)
+            shrd_blk = Block.deserialize(proof[0].begin_parse())
+            prev_blk = shrd_blk.info.prev_ref.prev
+
+            last_shard_blk = BlockIdExt.from_dict(prev_blk.__dict__ | {'workchain': last_shard_blk.workchain, 'shard': last_shard_blk.shard})
+            if last_shard_blk == blk:
+                return
+            raise LiteClientError('incorrect proof')
 
     async def raw_send_message(self, message: bytes):
         data = {'body': message}
@@ -786,7 +878,7 @@ class LiteClient:
         return result['status']
 
     @classmethod
-    def from_config(cls, config: dict, ls_i: int = 0, trust_level: int = 2):
+    def from_config(cls, config: dict, ls_i: int = 0, trust_level: int = 2, timeout: int = 10):
         ls = config['liteservers'][ls_i]
         init_block = config['validator']['init_block']
         init_block['file_hash'] = base64.b64decode(init_block['file_hash']).hex()
@@ -804,11 +896,11 @@ class LiteClient:
         )
 
     @classmethod
-    def from_mainnet_config(cls, ls_i: int = 0, trust_level: int = 0):
+    def from_mainnet_config(cls, ls_i: int = 0, trust_level: int = 0, timeout: int = 10):
         config = requests.get('https://ton.org/global-config.json').json()
-        return cls.from_config(config, ls_i, trust_level)
+        return cls.from_config(config, ls_i, trust_level, timeout)
 
     @classmethod
-    def from_testnet_config(cls, ls_i: int = 0, trust_level: int = 0):
+    def from_testnet_config(cls, ls_i: int = 0, trust_level: int = 0, timeout: int = 10):
         config = requests.get('https://ton.org/testnet-global.config.json').json()
-        return cls.from_config(config, ls_i, trust_level)
+        return cls.from_config(config, ls_i, trust_level, timeout)
